/// \file ambconv.h Main project header file

#ifndef __AMBCONV_H__
#define __AMBCONV_H__

/// maximum number of characters per line in AMBER topology file
#define MAX_CHAR_LINE 80 

/// main AMBER topology data structure
typedef struct
  {
    /// AMBER topology file title
    char title[MAX_CHAR_LINE];

    /** this variable is not included in AMBER topology file.
    *   If 'notitle' is equal to 0, in AMBER topology file is no title!
    */
    unsigned int notitle;

    /** \addtogroup AMBER_top_first_line */
    /*\@{*/

    /// number of atoms
    unsigned int natom;

    /// number of atom types
    unsigned int ntypes;

    /// bond w/o hydrogens
    unsigned int nbonh, mbona;

    /// angles w/o hydrogens
    unsigned int ntheth, mtheta;

    /// dihedrals w/o hydrogens
    unsigned int nphih, mphia;

    /// not used!
    unsigned int nhparm, nparm;

    /// number of excluded atoms
    unsigned int next;

    /// number of residues
    unsigned int nres;
    /*\@}*/

    /** \addtogroup AMBER_top_second_line */
    /*\@{*/

    /// bond,amgles,dihedrals without H + constrained bonds,angles,dihedrals
    unsigned int nbona, ntheta, nphia;

    /// unique types of bonds,angles,dihedrals
    unsigned int numbnd, numang, nptra;

    /// number of types in parameter file
    unsigned int natyp;

    /// number of 10-12 H-bond types (not used in new force fields)
    unsigned int nphb;

    /// perturbation data=1
    unsigned int ifpert;

    /// number of perturbed bons,angles,dihedrals
    unsigned int nbper, ngper, ndper;
    /*\@}*/

    /** \addtogroup AMBER_top_third_line */
    /*\@{*/

    /// number of bonds,angles,dihedrals in complete pert group ???
    unsigned int mbper, mgper, mdper;

    /// if periodic box is present, is set to 1
    unsigned int ifbox;

    /// number of atoms in largest residuum
    unsigned int nmxrs;

    /// cap=1
    unsigned int ifcap;

    /// polarization calculation
    unsigned int ipol;
    /*\@}*/

    /// Dynamically allocated data arrays
    /// user atom names (NATOM)
    char* igraph;

    /// atom charges (NATOM)
    double* chrg;

    /// atom masses (NATOM)
    double* amass;

    /// index for the atom types involved in L-J (NATOM)
    unsigned int* iac;

    /// number of excluded atom for atom i (NATOM)
    unsigned int* numex;

    /// index to arrays cn1,cn2, asol, bsol (NTYPES^2)
    unsigned int* ico;

    /// residue labels (NRES)
    char* labres;

    /// atom in residue (from ipres(i) to ipres(i+1)-1) ??? (NRES)
    unsigned int* ipres;

    /// bond force constant (NUMBND)
    double* rk;

    /// bond length (NUMBND)
    double* req;

    /// angle force constant (NUMANG)
    double* tk;

    /// angle value (NUMANG)
    double* teq;

    /// dihedral force constant (NPTRA)
    double* pk;

    /// dihedral periodicity (NPTRA)
    double* pn;

    /// dihedral phase (NPTRA)
    double* phase;

    /// unused (NATYP)
    double* solty;

    /// index to L-J ^12  ((NTYPES*(NTYPES+1))/2)
    double* cn1;

    /// index to L-J ^6  ((NTYPES*(NTYPES+1))/2)
    double* cn2;

    /// atoms in bonds with hydrogen (NBONH) -first atom
    unsigned int* ibh;
    /// atoms in bonds with hydrogen (NBONH) -second atom
    unsigned int* jbh;
    /// atoms in bonds with hydrogen (NBONH) -index
    unsigned int* icbh;

    /// atoms in bonds without hydrogen (MBONA) -first atom
    unsigned int* ib;
    /// atoms in bonds without hydrogen (MBONA) -second atom
    unsigned int* jb;
    /// atoms in bonds without hydrogen (MBONA) -index
    unsigned int* icb;

    /// atoms in angles with hydrogen (NTHETH) -first atom
    unsigned int* ith;
    /// atoms in angles with hydrogen (NTHETH) -second atom
    unsigned int* jth;
    /// atoms in angles with hydrogen (NTHETH) -third atom
    unsigned int* kth;
    /// atoms in angles with hydrogen (NTHETH) -index
    unsigned int* ickh;

    /// atoms in angles without hydrogen (MTHETA) -first atom
    unsigned int* it;
    /// atoms in angles without hydrogen (MTHETA) -second atom
    unsigned int* jt;
    /// atoms in angles without hydrogen (MTHETA) -third atom
    unsigned int* kt;
    /// atoms in angles without hydrogen (MTHETA) -index
    unsigned int* ict;


    int* iph; /// atoms in dihedrals with hydrogen (NPHIH)
    int* jph;
    int* kph;
    int* lph;
    unsigned int* icph;
    int* ip; /// atoms in dihedrals without hydrogen (MPHIA)
    int* jp;
    int* kp;
    int* lp;
    unsigned int* icp;
    unsigned int* natex; /// excluded atom list (NEXT)
    double* asol; /// r^12 index for H-bonds -not used in parm96 (NPHB)
    double* bsol; /// r^10 index for H-bonds - dtto -
    double* hbcut; /// not used (NPHB)
    char* isymbl;  /// AMBER atom types (NATOM)
    char* itree;   /// tree joining information (NATOM)
    int* join;    /// ancient tree joining info, not used (NATOM)
    unsigned int* irotat; /// unused (NATOM)
    /// only if IFBOX > 0
    unsigned int iptres; /// not used
    unsigned int nspm; /// total number of molecules
    unsigned int nspsol; /// first solvent molecule
    unsigned int* nsp; /// total number of atoms in each molecule (NSPM)
    double beta, boxx, boxy, boxz; /// periodic box parameters
    /// only if IFCAP > 0
    unsigned int natcap; /// last atom before start of the 'cap'
    double cutcap, xcap, ycap, zcap; /// 'cap' parameters
    /// only if IFPERT > 0
    /// initialstate = prep/link/edit = 1, finalstate = parm = 0
    unsigned int* ibper; /// atoms in perturbed bonds (NBPER)
    unsigned int* jbper;    
    unsigned int* icbper; /// pointer into rk, req. icbper(i)==lambda1,
                            /// icbper(i+nbper)==lambda0 (NBPER*2)
    unsigned int* ipter; /// atoms in perturbed angles (NGPER)
    unsigned int* jpter;    
    unsigned int* kpter;
    unsigned int* icpter; /// pointer into rk, req. icpter(i)==lambda1,
                            /// icpter(i+ngper)==lambda0 (NGPER*2)
    int* ipper; /// atoms in perturbed dihedrals (NDPER)
    int* jpper;
    int* kpper;
    int* lpper;
    unsigned int* icpper; /// pointer into arrays pk, pn and phase (NDPER*2)

    char* labresper; /// residue names at lambda=0 (NRES)
    char* igrper; /// atom names at 0 (NATOM)
    char* ismper; /// atomic symbols at 0 (NATOM)
    double* almper; /// !!!!! atom masses at lambda=0 (NATOM) !!!!!  p.s. PARMF
    unsigned int* iaper; /// is 1 if atom is being perturbed (NATOM)
    unsigned int* iacper; /// index to LJ arrays at 0, similar to IAC (NATOM)
    double* cgper; /// charges at 0 (NATOM)
   
    /// IPOL is not in AMBER topology documentation
    /// only if IPOL == 1
    double* atpol; /// atomic polarizabilities (NATOM)
    /// only if IPOL == 1 and IFPERT == 1
    double* atpol1; /// atomic polarizabilites at lambda=1 (above at lambda=0) (NATOM)
} s_atop;

/// main functions
void read_ambertop(s_atop* atop); /// read AMBER topology data to struct s_atop
void write_itp(s_atop* atop); /// write GROMACS forcefield ITP+NB+ATP files
void write_pdb(s_atop* atop); /// write pdb file
void write_grotop(s_atop* atop); /// write residue database
void write_gro(s_atop* atop); /// write GROMACS .top
void read_rst(s_atop* atop); ///
void write_constraints(s_atop* atop); ///
void write_lacc_rst(s_atop* atop);
void write_position_constraints(s_atop* atop); // [ position_constraints ]
void write_tintop(s_atop* atop); // write TINKER parameter=topology file
void write_tintop_bon(s_atop* atop); // write TINKER bonded parameters
void write_xyz(s_atop* atop); // write TINKER xyz coordinates
void write_key(s_atop* atop); // write TINKER key file
void write_tintop_charge(s_atop* atop); // write TINKER charges to .prm file

char* ambtop_filename; /**< AMBER topology */
char* grotop_filename; /**< GROMACS topology */
char* rst_filename; /// AMBER coordinates
char* gro_filename; /// GROMACS coordinates
char* pdb_filename;
char* common_filename;
char* groitp_filename;
char* gronb_filename;
char* groatp_filename;
char* sol_filename; /**< solvent filename */
char* lacc_rst_filename; /* .rst coordinates with low accuracy */
char* constraints_filename;
char* tintop_filename; // TINKER parameters
char* xyz_filename; // TINKER coordinates
char* key_filename; // TINKER key file

/// file descriptor
FILE* ambtop;
FILE* grotop;
FILE* constraints_file;
FILE* tintop;

/// global parameters
double* coords;
unsigned long int ncoords;
double box[3];
char rst_title[MAX_CHAR_LINE];
unsigned int sol;   /// if set -sol "solvent_name", include solvent_itp_filename to topology
double conv; /// conversion factor cal -> J  4.1868
unsigned int nbpairs; /// if set -nbpairs, print nonbonded pairs to ff???nb.itp
unsigned int constraints; /// if -const set, add section [constraints] before [pairs]
unsigned int lacc_rst; // if -lacc_rst is set, write low accuracy coorinets out
unsigned int position_constraints; // flag -prest 
unsigned int pdb; // control variables, see ambconv.c
unsigned int tinker;
unsigned int gromacs;
unsigned int invert; // swap free energy states A and B

/// service functions
/// Close 'file'
void close_file(FILE* file);

/// Open file 'filename' in read(<B>r</B>) or write(<B>w</B>) mode
FILE* open_file(char* filename, char* mode);

/// System error code
extern int errno;

/// Print error error message
void printerror(char* message);

/// Read and parse command line
void read_commandline(int argc, char* argv[]);

/// Read string from top
void readstring(unsigned int length, char* source);

/// Extended version of malloc
void* xmalloc(size_t size);

/// Extended version of free
void xfree(void* mem);

/// Return 1 if 'atom' is between 'from' and 'to'
unsigned int out(int atom, int from, int to);

/**
* Return new string from the 'input' string.
  String starts at 'distance', length is 'length'
*/
//char* get_slice(char* input, unsigned int length, unsigned int distance, unsigned int index);
char* get_slice(char* input, char* output,unsigned int length, unsigned int distance, unsigned int index);

/// find ekvivalent types
unsigned int unique(char** types, char* type, unsigned int i);

/// if fabs(num1-num2) > limit --> return 1
unsigned int not_ekviv(double num1, double num2, double limit);

/// if num1 != num2  --> return 1
unsigned int i_not_ekviv(int num1, int num2);

/// return bond length between atom1 and atom2
double bond_length(s_atop* atop, int atom1, int atom2);

/// return distance between 'a' and 'c' in angle a-b-c
double dist_angle13(double rab, double rbc, double cosine);

/**
* Replace apostrophe in input_string with "q".
  Function searches 'input_string' by 'strrchr' function
  and replace each occurence of the apostrophe
*/
char* repapo(char* input_string);

double version; // version number

#endif
