/* AMBGRO - program for convering AMBER topology to GROMACS topology */
/* Filip Ryjacek 2001 */

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <stddef.h>
#include <string.h>
#include <math.h>
#include <mcheck.h>

#include "ambconv.h"


void read_ambertop(s_atop* atop)
{
  char* retchar;
  int i, tmp;
  int lambda;

/* debug memory allocation */
#ifdef DEBUG_MEM
  mtrace();
#endif

/* open AMBER topology file */
ambtop = open_file(ambtop_filename, "r");

/* read title from AMBER top */
if ((retchar=fgets(atop->title, MAX_CHAR_LINE, ambtop)) == NULL) printerror("Title not read");
  atop->title[80] = '\0';
#ifdef DEBUG_ATOP
  printf("\nTitle: %s\n", atop->title);
  printf("Title length: %d\n", strlen(atop->title));
#endif

/* set notitle to prevent long empty lines in file */
  if (strlen(atop->title) == 79)
    {
      // has title
      atop->notitle = 0;
    }
  else
    {
      atop->notitle = 1;
    }

/* AMBER topology header */
/* read first line */
if (fscanf(ambtop, "%6d%6d%6d%6d%6d%6d%6d%6d%6d%6d%6d%6d", 
  &atop->natom, &atop->ntypes, &atop->nbonh, &atop->mbona, &atop->ntheth, &atop->mtheta, 
  &atop->nphih, &atop->mphia, &atop->nhparm, &atop->nparm, &atop->next, &atop->nres) !=12) 
  printerror("AMBER TOP: too few parameters on second line");

/* read second line */
if (fscanf(ambtop, "%6d%6d%6d%6d%6d%6d%6d%6d%6d%6d%6d%6d",
  &atop->nbona, &atop->ntheta, &atop->nphia, &atop->numbnd, &atop->numang, &atop->nptra, 
  &atop->natyp, &atop->nphb, &atop->ifpert, &atop->nbper, &atop->ngper, &atop->ndper) !=12)
  printerror("AMBER TOP: too few parameters on third line");

/* read third line */
if (fscanf(ambtop, "%6d%6d%6d%6d%6d%6d",
  &atop->mbper, &atop->mgper, &atop->mdper, &atop->ifbox, &atop->nmxrs, &atop->ifcap) !=6)
  printerror("AMBER TOP: too few parameters on fourth line");
/* end of AMBER topology header */


/* allocate dynamic data arrays */
/* ATOM NAMES */
getc(ambtop); /* remove '\n' from buffer */
atop->igraph = (char*) xmalloc((atop->natom)*4+1);
readstring((atop->natom)*4, atop->igraph);
#ifdef DEBUG_ATOP
printf("\natom names (%d)\n|%s|\n", atop->natom, atop->igraph); 
#endif

/* CHARGE */
atop->chrg = (double*) xmalloc((sizeof(double))*(atop->natom));
#ifdef DEBUG_ATOP
printf("\ncharges:\n");
#endif
for(i=0; i<atop->natom; i++)
  {
    fscanf(ambtop, " %le", atop->chrg+i);
    atop->chrg[i] = atop->chrg[i]/sqrt(332.0000);  /* convert to electrons */
#ifdef DEBUG_ATOP
    printf("%d\t% .6f\n", i, atop->chrg[i]);
#endif
  }

/* ATOM MASSES */
atop->amass = (double*) xmalloc((sizeof(double))*(atop->natom));
#ifdef DEBUG_ATOP
printf("\nmasses:\n");
#endif
for(i=0; i<atop->natom; i++)
  {
    fscanf(ambtop, " %le", atop->amass+i);
#ifdef DEBUG_ATOP
    printf("%d\t% 8.4f\n", i, atop->amass[i]); 
#endif
  }

/* INDEX TO L-J ARRAYS */
atop->iac = (unsigned int*) xmalloc((sizeof(unsigned int))*(atop->natom));
#ifdef DEBUG_ATOP
printf("\nIAC L-J indexes:\n");
#endif
for(i=0; i<atop->natom; i++)
  {
    fscanf(ambtop, " %d", atop->iac+i);
#ifdef DEBUG_ATOP
    printf("%d\t% d\n", i, atop->iac[i]);
#endif
  }

/* NUMBER OF EXCLUDED ATOMS */
atop->numex = (unsigned int*) xmalloc((sizeof(unsigned int))*(atop->natom));
#ifdef DEBUG_ATOP
printf("\nnumber of excluded atoms:\n");
#endif
for(i=0; i<atop->natom; i++)
  {
    fscanf(ambtop, " %d", atop->numex+i);
#ifdef DEBUG_ATOP
    printf("%d\t% d\n", i, atop->numex[i]); 
#endif
  }

/* INDEX TO ARRAYS CN1, CN2 */
atop->ico = (unsigned int*) xmalloc(sizeof(unsigned int)*(atop->ntypes)*(atop->ntypes));
#ifdef DEBUG_ATOP
printf("\nindexes to arrays cn1, cn2:\n");
#endif
for(i=0; i<(atop->ntypes)*(atop->ntypes); i++)
  {
    fscanf(ambtop, " %d", atop->ico+i);
#ifdef DEBUG_ATOP
    printf("%d\t% d\n", i, atop->ico[i]); 
#endif
  }

/* RESIDUE NAMES */
getc(ambtop); /* remove '\n' from buffer */
atop->labres = (char*) xmalloc((atop->nres)*4+1);
readstring((atop->nres)*4, atop->labres);
#ifdef DEBUG_ATOP
printf("\nresidue names\n|%s|\n", atop->labres);
#endif

/* ATOM IN RESIDUE ??? */
atop->ipres = (unsigned int*) xmalloc((sizeof(unsigned int))*(atop->nres));
#ifdef DEBUG_ATOP
printf("\natom in residue ???:\n");
#endif
for(i=0; i<atop->nres; i++)
  {
    fscanf(ambtop, " %d", atop->ipres+i);
#ifdef DEBUG_ATOP
    printf("%d\t% d\n", i, atop->ipres[i]); 
#endif
  }

/* BOND FORCE CONSTANT */
atop->rk = (double*) xmalloc((sizeof(double))*(atop->numbnd));
#ifdef DEBUG_ATOP
printf("\nbond force constant (kcal/mol):\n");
#endif
for(i=0; i<atop->numbnd; i++)
  {
    fscanf(ambtop, " %le", atop->rk+i);
#ifdef DEBUG_ATOP
    printf("%d\t% .4f\n", i, atop->rk[i]); 
#endif
  }

/* EQUILIBRIUM BOND LENGTH */
atop->req = (double*) xmalloc((sizeof(double))*(atop->numbnd));
#ifdef DEBUG_ATOP
printf("\nequilibrium bond length (A):\n");
#endif
for(i=0; i<atop->numbnd; i++)
  {
    fscanf(ambtop, " %le", atop->req+i);
#ifdef DEBUG_ATOP
    printf("%d\t% .4f\n", i, atop->req[i]);
#endif
  }

/* ANGLE FORCE CONSTANT */
atop->tk = (double*) xmalloc((sizeof(double))*(atop->numang));
#ifdef DEBUG_ATOP
printf("\nangle force constant (kcal/mol.deg^2):\n");
#endif
for(i=0; i<atop->numang; i++)
  {
    fscanf(ambtop, " %le", atop->tk+i);
#ifdef DEBUG_ATOP
    printf("%d\t% 7.2f\n", i, atop->tk[i]); 
#endif
  }

/* EQUILIBRIUM ANGLE VALUES */
atop->teq = (double*) xmalloc((sizeof(double))*(atop->numang));
#ifdef DEBUG_ATOP
printf("\nequilibrium angle values (deg.):\n");
#endif
for(i=0; i<atop->numang; i++)
  {
    fscanf(ambtop, " %le", atop->teq+i);
    atop->teq[i] = atop->teq[i]*57.29577951;
#ifdef DEBUG_ATOP
    printf("%d\t% .2f\n", i, atop->teq[i]);
#endif
  }

/* DIHEDRAL FORCE CONSTANT */
atop->pk = (double*) xmalloc((sizeof(double))*(atop->nptra));
#ifdef DEBUG_ATOP
printf("\ndihedral force constant (kcal/mol.deg^2):\n");
#endif
for(i=0; i<atop->nptra; i++)
  {
    fscanf(ambtop, " %le", atop->pk+i);
#ifdef DEBUG_ATOP
    printf("%d\t% 7.2f\n", i, atop->pk[i]); 
#endif
  }

/* DIHEDRAL PERIODICITY */
atop->pn = (double*) xmalloc((sizeof(double))*(atop->nptra));
#ifdef DEBUG_ATOP
printf("\ndihedral periodicity:\n");
#endif
for(i=0; i<atop->nptra; i++)
  {
    fscanf(ambtop, " %le", atop->pn+i);
#ifdef DEBUG_ATOP
    printf("%d\t% .0f\n", i, atop->pn[i]); 
#endif
  }

/* DIHEDRAL PHASE */
atop->phase = (double*) xmalloc((sizeof(double))*(atop->nptra));
#ifdef DEBUG_ATOP
printf("\ndihedral phase (deg.):\n");
#endif
for(i=0; i<atop->nptra; i++)
  {
    fscanf(ambtop, " %le", atop->phase+i);
    atop->phase[i] = atop->phase[i]*57.29577951;
#ifdef DEBUG_ATOP
    printf("%d\t% .2f\n", i, atop->phase[i]); 
#endif
  }

/* SOLTY ARRAY - not used but present */
atop->solty = (double*) xmalloc((sizeof(double))*(atop->natyp));
#ifdef DEBUG_ATOP
printf("\nsolty (not used):\n");
#endif
for(i=0; i<atop->natyp; i++)
  {
    fscanf(ambtop, " %le", atop->solty+i);
/*  printf("%d\t% .2f\n", i, atop->phase[i]); */
  }

/* CN1 - LJ r^12 */
atop->cn1 = (double*) xmalloc((sizeof(double))*((atop->ntypes)*(atop->ntypes+1)/2));
#ifdef DEBUG_ATOP
printf("\nLJ r^12:\n");
#endif
for(i=0; i<((atop->ntypes)*(atop->ntypes+1)/2); i++)
  {
    fscanf(ambtop, " %le", atop->cn1+i);
#ifdef DEBUG_ATOP
    printf("%d\t% 16.8f\n", i, atop->cn1[i]); 
#endif
  }

/* CN2 - LJ r^6 */
atop->cn2 = (double*) xmalloc((sizeof(double))*((atop->ntypes)*(atop->ntypes+1)/2));
#ifdef DEBUG_ATOP
printf("\nLJ r^6:\n");
#endif
for(i=0; i<((atop->ntypes)*(atop->ntypes+1)/2); i++)
  {
    fscanf(ambtop, " %le", atop->cn2+i);
#ifdef DEBUG_ATOP
    printf("%d\t% 16.8f\n", i, atop->cn2[i]); 
#endif
  }

/* ATOMS IN BONDS (WITH AND WITHOUT HYDROGEN) */
atop->ibh = (unsigned int*) xmalloc(sizeof(unsigned int)*(atop->nbonh));
atop->jbh = (unsigned int*) xmalloc(sizeof(unsigned int)*(atop->nbonh));
atop->icbh =(unsigned int*) xmalloc(sizeof(unsigned int)*(atop->nbonh));
#ifdef DEBUG_ATOP
printf("\natoms in bonds:\n");
#endif
for(i=0; i<atop->nbonh; i++)
  {
    fscanf(ambtop, "%6d%6d%6d", atop->ibh+i, atop->jbh+i, atop->icbh+i);
    atop->ibh[i] = (atop->ibh[i]/3)+1; /* convert to atom numbers */
    atop->jbh[i] = (atop->jbh[i]/3)+1;
#ifdef DEBUG_ATOP
    printf("%d:\t% d\t% d\t% d\n", i, atop->ibh[i], atop->jbh[i], (atop->icbh[i])-1); 
#endif
  }

atop->ib = (unsigned int*) xmalloc(sizeof(unsigned int)*(atop->mbona));
atop->jb = (unsigned int*) xmalloc(sizeof(unsigned int)*(atop->mbona));
atop->icb =(unsigned int*) xmalloc(sizeof(unsigned int)*(atop->mbona));
for(i=0; i<atop->mbona; i++)
  {
    fscanf(ambtop, "%6d%6d%6d", atop->ib+i, atop->jb+i, atop->icb+i);
    atop->ib[i] = (atop->ib[i]/3)+1; /* convert to atom numbers */
    atop->jb[i] = (atop->jb[i]/3)+1;
#ifdef DEBUG_ATOP
    printf("%d:\t% d\t% d\t% d\n", i, atop->ib[i], atop->jb[i], (atop->icb[i])-1); 
#endif
  }

/* ATOMS IN ANGLES (WITH AND WITHOUT HYDROGEN) */
atop->ith = (unsigned int*) xmalloc((sizeof(unsigned int))*(atop->ntheth));
atop->jth = (unsigned int*) xmalloc((sizeof(unsigned int))*(atop->ntheth));
atop->kth = (unsigned int*) xmalloc((sizeof(unsigned int))*(atop->ntheth));
atop->ickh = (unsigned int*) xmalloc((sizeof(unsigned int))*(atop->ntheth));
#ifdef DEBUG_ATOP
printf("\natoms in angles:\n");
#endif
for(i=0; i<atop->ntheth; i++)
  {
    fscanf(ambtop, "%6d%6d%6d%6d", atop->ith+i,atop->jth+i,atop->kth+i,atop->ickh+i);
    atop->ith[i] = (atop->ith[i]/3)+1; /* convert to atom numbers */
    atop->jth[i] = (atop->jth[i]/3)+1;
    atop->kth[i] = (atop->kth[i]/3)+1;
#ifdef DEBUG_ATOP
    printf("%d:\t% d\t% d\t% d\t% d\n", i,atop->ith[i],atop->jth[i],atop->kth[i],(atop->ickh[i])-1); 
#endif
  }

atop->it = (unsigned int*) xmalloc((sizeof(unsigned int))*(atop->mtheta));
atop->jt = (unsigned int*) xmalloc((sizeof(unsigned int))*(atop->mtheta));
atop->kt = (unsigned int*) xmalloc((sizeof(unsigned int))*(atop->mtheta));
atop->ict = (unsigned int*) xmalloc((sizeof(unsigned int))*(atop->mtheta));
for(i=0; i<atop->mtheta; i++)
  {
    fscanf(ambtop, "%6d%6d%6d%6d", atop->it+i,atop->jt+i,atop->kt+i,atop->ict+i);
    atop->it[i] = (atop->it[i]/3)+1; /* convert to atom numbers */
    atop->jt[i] = (atop->jt[i]/3)+1;
    atop->kt[i] = (atop->kt[i]/3)+1;
#ifdef DEBUG_ATOP
    printf("%d:\t% d\t% d\t% d\t% d\n", i,atop->it[i],atop->jt[i],atop->kt[i],(atop->ict[i])-1); 
#endif
  }

/* ATOMS IN DIHEDRALS (WITH AND WITHOUT HYDROGEN) */
atop->iph = (int*) xmalloc((sizeof(int))*(atop->nphih));
atop->jph = (int*) xmalloc((sizeof(int))*(atop->nphih));
atop->kph = (int*) xmalloc((sizeof(int))*(atop->nphih));
atop->lph = (int*) xmalloc((sizeof(int))*(atop->nphih));
atop->icph = (unsigned int*) xmalloc((sizeof(unsigned int))*(atop->nphih));
#ifdef DEBUG_ATOP
printf("\natoms in dihedrals:\n");
#endif
for(i=0; i<atop->nphih; i++)
  {
    fscanf(ambtop, "%6d%6d%6d%6d%6d", atop->iph+i,atop->jph+i,atop->kph+i,atop->lph+i,atop->icph+i);
    atop->iph[i] = (atop->iph[i]/3)+1; /* convert to atom numbers */
    atop->jph[i] = (atop->jph[i]/3)+1;
    if (atop->kph[i] < 0) /* end group */
      {
        atop->kph[i] = -((abs(atop->kph[i])/3)+1);
      }
    else
      {
        atop->kph[i] = (atop->kph[i]/3)+1;
      }
    if (atop->lph[i] < 0) /* improper dihedral */
      {
        atop->lph[i] = -((abs(atop->lph[i])/3)+1);
      }
    else
      {
        atop->lph[i] = (atop->lph[i]/3)+1;
      }
#ifdef DEBUG_ATOP
    printf("%d:\t% d\t% d\t% d\t% d\t#% d\n", i,atop->iph[i],atop->jph[i],atop->kph[i],atop->lph[i],(atop->icph[i])-1); 
#endif
  }

atop->ip = (int*) xmalloc((sizeof(int))*(atop->mphia));
atop->jp = (int*) xmalloc((sizeof(int))*(atop->mphia));
atop->kp = (int*) xmalloc((sizeof(int))*(atop->mphia));
atop->lp = (int*) xmalloc((sizeof(int))*(atop->mphia));
atop->icp = (unsigned int*) xmalloc((sizeof(unsigned int))*(atop->mphia));
for(i=0; i<atop->mphia; i++)
  {
    fscanf(ambtop, "%6d%6d%6d%6d%6d", atop->ip+i,atop->jp+i,atop->kp+i,atop->lp+i,atop->icp+i);
    atop->ip[i] = (atop->ip[i]/3)+1; /* convert to atom numbers */
    atop->jp[i] = (atop->jp[i]/3)+1;
    if (atop->kp[i] < 0) /* end group */
      {
        atop->kp[i] = -((abs(atop->kp[i])/3)+1);
      }
    else
      {
        atop->kp[i] = (atop->kp[i]/3)+1;
      }
    if (atop->lp[i] < 0) /* improper dihedral */
      {
        atop->lp[i] = -((abs(atop->lp[i])/3)+1);
      }
    else
      {
        atop->lp[i] = (atop->lp[i]/3)+1;
      }
#ifdef DEBUG_ATOP
    printf("%d:\t% d\t% d\t% d\t% d\t#% d\n", i,atop->ip[i],atop->jp[i],atop->kp[i],atop->lp[i],(atop->icp[i])-1); 
#endif
  }

/* EXCLUDED ATOM LIST */
atop->natex = (unsigned int*) xmalloc((sizeof(unsigned int))*(atop->next));
#ifdef DEBUG_ATOP
printf("\nexcluded atom list:\n");
#endif
for(i=0; i<atop->next; i++)
  {
    fscanf(ambtop, " %d", atop->natex+i);
#ifdef DEBUG_ATOP
    printf("%d\t% d\n", i, atop->natex[i]); 
#endif
  }

if (atop->nphb ==0)
  {
#ifdef DEBUG_ATOP
    printf("\nnphb: %d\t Skipping 3 lines...\n", atop->nphb);
#endif
    getc(ambtop);
    getc(ambtop);
    getc(ambtop);
  }
else
  {

#ifdef DEBUG_ATOP
  printf("\nThree unused (but still present) arrays for explicit H-bonds\n");
#endif
/* r^12 index for H-bonds */
atop->asol = (double*) xmalloc((sizeof(double))*(atop->nphb));
for(i=0; i<atop->nphb; i++)
  {
    fscanf(ambtop, " %le", atop->asol+i);
#ifdef DEBUG_ATOP
    printf("%d\t% 7.2f\n", i, atop->asol[i]); 
#endif
  }

/* r^10 index for H-bonds */
atop->bsol = (double*) xmalloc((sizeof(double))*(atop->nphb));
for(i=0; i<atop->nphb; i++)
  {
    fscanf(ambtop, " %le", atop->bsol+i);
#ifdef DEBUG_ATOP
    printf("%d\t% 7.2f\n", i, atop->bsol[i]); 
#endif
  }

/* something for H-bonds - not used */
atop->hbcut = (double*) xmalloc((sizeof(double))*(atop->nphb));
for(i=0; i<atop->nphb; i++)
  {
    fscanf(ambtop, " %le", atop->hbcut+i);
#ifdef DEBUG_ATOP
    printf("%d\t% 7.2f\n", i, atop->hbcut[i]); 
#endif
  }
  } /* end if nphb == 0 */

/* AMBER ATOM TYPES */
getc(ambtop); /* remove '\n' from buffer */
atop->isymbl = (char*) xmalloc((atop->natom)*4+1);
readstring((atop->natom)*4, atop->isymbl);
#ifdef DEBUG_ATOP
printf("\natom types\n|%s|\n", atop->isymbl);
#endif

/* TREE JOINING INFORMATION */
getc(ambtop); /* remove '\n' from buffer */
atop->itree = (char*) xmalloc((atop->natom)*4+1);
readstring((atop->natom)*4, atop->itree);
#ifdef DEBUG_ATOP
printf("\ntree joining info\n|%s|\n", atop->itree); 
#endif

/* ANCIENT TREE INFO - not used */
atop->join = (int*) xmalloc((sizeof(int))*(atop->natom));
#ifdef DEBUG_ATOP
printf("\nancient tree info - not used\n");
#endif
for(i=0; i<atop->natom; i++)
  {
    fscanf(ambtop, " %d", atop->join+i);
#ifdef DEBUG_ATOP
/*  printf("%d\t% d\n", i, atop->join[i]); */
#endif
  }

/* ??? - not used */
atop->irotat = (unsigned int*) xmalloc((sizeof(unsigned int))*(atop->natom));
#ifdef DEBUG_ATOP
printf("\nirotat - not used\n");
#endif
for(i=0; i<atop->natom; i++)
  {
    fscanf(ambtop, " %d", atop->irotat+i);
#ifdef DEBUG_ATOP
/*  printf("%d\t% d\n", i, atop->irotat[i]); */
#endif
  }

/* ------------------------------BOX--------------------------------- */
if (atop->ifbox > 0)
  {
#ifdef DEBUG_ATOP
    printf("\nreading first part of BOX info...\n");
#endif 
  if (fscanf(ambtop, "%6d%6d%6d", &atop->iptres, &atop->nspm, &atop->nspsol) !=3)
    {
      printerror("AMBER TOP: too few parameters in first line of BOX info");
    }
  atop->nsp = (unsigned int*) xmalloc((sizeof(unsigned int))*(atop->nres));
#ifdef DEBUG_ATOP
  printf("\nreading second part of BOX info...\n");
  printf("nsp: number of atoms in each residue\n");
#endif
  for(i=0; i<atop->nres; i++)
    {
      fscanf(ambtop, " %d", atop->nsp+i);
#ifdef DEBUG_ATOP
  printf("%d\t% d\n", i, atop->nsp[i]);
#endif
    }
#ifdef DEBUG_ATOP
    printf("\nreading third part of BOX info...\n");
#endif 
  if (fscanf(ambtop, " %le %le %le %le", &atop->beta, &atop->boxx, &atop->boxy, &atop->boxz) !=4)
    {
      printerror("AMBER TOP: cannot read BOX dimensions");
    }
  }
else
  {
#ifdef DEBUG_ATOP
  printf("\nno BOX info\n");
#endif
  }

/* ------------------------------CAP--------------------------------- */
if (atop->ifcap > 0)
  {
#ifdef DEBUG_ATOP
  printf("\nreading CAP info...\n");
#endif
  if (fscanf(ambtop, "%6d", &atop->natcap) !=1)
    {
      printerror("AMBER TOP: cannot read first-1 CAP atom (1st line)");
    }
  if (fscanf(ambtop, " %le %le %le %le", &atop->cutcap, &atop->xcap, &atop->ycap, &atop->zcap) !=4)
    {
      printerror("AMBER TOP: cannot read CAP parameters (2nd line)");
    }
  }
else
  {
#ifdef DEBUG_ATOP
  printf("\nno CAP info\n");
#endif
  }

/* ------------------------------PERTURBATION------------------------ */
if (atop->ifpert > 0)
  {
#ifdef DEBUG_ATOP
  printf("\nreading PERTURBATION info...\n");
#endif

/* PERTURBED ATOMS IN BONDS */
atop->ibper =  (unsigned int*) xmalloc(sizeof(unsigned int)*(atop->nbper));
atop->jbper =  (unsigned int*) xmalloc(sizeof(unsigned int)*(atop->nbper));
atop->icbper = (unsigned int*) xmalloc(sizeof(unsigned int)*(atop->nbper)*2);

#ifdef DEBUG_ATOP
 printf("\nPERT: natoms in bonds:\n");
#endif

for(i=0; i<atop->nbper; i++)
  {
    fscanf(ambtop, "%6d%6d", atop->ibper+i, atop->jbper+i);
    atop->ibper[i] = (atop->ibper[i]/3)+1;  /* convert to atom numbers */
    atop->jbper[i] = (atop->jbper[i]/3)+1;
  }

for(i=0; i < (atop->nbper*2); i++)
  {
    fscanf(ambtop, "%6d", atop->icbper+i);
    tmp = i;
    lambda = 1;
#ifdef DEBUG_ATOP
    if (i >= atop->nbper) { tmp= i - atop->nbper; lambda = 0; }
    printf("%6d:%8d%8d\t|%4d    l=%d\n", tmp, atop->ibper[tmp], atop->jbper[tmp], (atop->icbper[i])-1, lambda); 
#endif
  }


/* PERTURBED ATOMS IN ANGLES */
atop->ipter = (unsigned int*) xmalloc((sizeof(unsigned int))*(atop->ngper));
atop->jpter = (unsigned int*) xmalloc((sizeof(unsigned int))*(atop->ngper));
atop->kpter = (unsigned int*) xmalloc((sizeof(unsigned int))*(atop->ngper));
atop->icpter = (unsigned int*) xmalloc((sizeof(unsigned int))*(atop->ngper)*2);

#ifdef DEBUG_ATOP
  printf("\nPER: atoms in angles:\n");
#endif

for(i=0; i<atop->ngper; i++)
  {
    fscanf(ambtop, "%6d%6d%6d", atop->ipter+i,atop->jpter+i,atop->kpter+i);
    atop->ipter[i] = (atop->ipter[i]/3)+1; /* convert to atom numbers */ 
    atop->jpter[i] = (atop->jpter[i]/3)+1;
    atop->kpter[i] = (atop->kpter[i]/3)+1;
  }

for(i=0; i < (atop->ngper)*2; i++)
  {
    fscanf(ambtop, "%6d", atop->icpter+i);
    tmp = i;
    lambda = 1;
#ifdef DEBUG_ATOP
    if (i >= atop->ngper) { tmp= i - atop->ngper; lambda = 0; }
    printf("%6d:%8d%8d%8d\t|%4d    l=%d\n", tmp,atop->ipter[tmp],atop->jpter[tmp],atop->kpter[tmp],(atop->icpter[i])-1, lambda); 
#endif
  }


/* PERTURBED ATOMS IN DIHEDRALS */
atop->ipper = (int*) xmalloc((sizeof(int))*(atop->ndper));
atop->jpper = (int*) xmalloc((sizeof(int))*(atop->ndper));
atop->kpper = (int*) xmalloc((sizeof(int))*(atop->ndper));
atop->lpper = (int*) xmalloc((sizeof(int))*(atop->ndper));
atop->icpper = (unsigned int*) xmalloc((sizeof(unsigned int))*(atop->ndper)*2);

#ifdef DEBUG_ATOP
  printf("\nPER: atoms in dihedrals:\n");
#endif

for(i=0; i<atop->ndper; i++)
  {
    fscanf(ambtop, "%6d%6d%6d%6d", atop->ipper+i,atop->jpper+i,atop->kpper+i,atop->lpper+i);
    /* convert to atom numbers */
    atop->ipper[i] = (atop->ipper[i]/3)+1;
    atop->jpper[i] = (atop->jpper[i]/3)+1;
    if (atop->kpper[i] < 0) /* end group interaction */
      {
        atop->kpper[i] = -((abs(atop->kpper[i])/3)+1);
      }
    else
      {
        atop->kpper[i] = (atop->kpper[i]/3)+1;
      }
    if (atop->lpper[i] < 0) /* improper dihedral */
      {
        atop->lpper[i] = -((abs(atop->lpper[i])/3)+1);
      }
    else
      {
        atop->lpper[i] = (atop->lpper[i]/3)+1;
      }
  }

for(i=0; i < (atop->ndper)*2; i++)
  {
    fscanf(ambtop, "%6d", atop->icpper+i);
    tmp = i;
    lambda = 1;
#ifdef DEBUG_ATOP
    if (i >= atop->ndper) { tmp= i - atop->ndper; lambda = 0; }
    printf("%6d:%8d%8d%8d%8d\t|%4d    l=%d\n", tmp, atop->ipper[tmp], atop->jpper[tmp], atop->kpper[tmp], atop->lpper[tmp], (atop->icpper[i])-1, lambda); 
#endif
  }


/* PERTURBATION RESIDUE NAMES */
getc(ambtop); /* remove '\n' from buffer */
atop->labresper = (char*) xmalloc((atop->nres)*4+1);
readstring((atop->nres)*4, atop->labresper);

#ifdef DEBUG_ATOP
  printf("\nPER: residue names\n|%s|\n", atop->labresper);
#endif

/* PERTURBATION ATOM NAMES */
getc(ambtop); /* remove '\n' from buffer */
atop->igrper = (char*) xmalloc((atop->natom)*4+1);
readstring((atop->natom)*4, atop->igrper);

#ifdef DEBUG_ATOP
  printf("\nPER: atom names (%d)\n|%s|\n", atop->natom, atop->igrper); 
#endif

/* AMBER ATOM TYPES */
getc(ambtop); /* remove '\n' from buffer */
atop->ismper = (char*) xmalloc((atop->natom)*4+1);
readstring((atop->natom)*4, atop->ismper);

#ifdef DEBUG_ATOP
  printf("\nPER: atom types\n|%s|\n", atop->ismper);
#endif


/* !!! ATOM MASSES AT LAMBDA=0 !!!   PARMF */
#ifdef DEBUG_ATOP
  printf("\nPER: atom masses at lambda=0:\n");
#endif
atop->almper = (double*) xmalloc((sizeof(double))*(atop->natom));
for(i=0; i < atop->natom; i++)
  {
    fscanf(ambtop, " %le", atop->almper+i);
#ifdef DEBUG_ATOP
    printf("%6d:%20.8f\n", i, atop->almper[i]); 
#endif
  }


/* IF IAPER=1, ATOM IS BEING PERTURBED */
atop->iaper = (unsigned int*) xmalloc((sizeof(int))*(atop->natom));
#ifdef DEBUG_ATOP
  printf("\nPER: if 1, atom is being perturbed:\n");
#endif
for(i=0; i<atop->natom; i++)
  {
    fscanf(ambtop, "%6d", atop->iaper+i);
#ifdef DEBUG_ATOP
    if ( atop->iaper[i] == 1)
      {
        printf("%6d:%10s\n", i, "is perturbed"); 
      }
    else
      {
        printf("%6d:\n", i); 
      }
#endif
  }


/* INDEX TO PERTURBED L-J ARRAYS */
atop->iacper = (unsigned int*) xmalloc((sizeof(unsigned int))*(atop->natom));
#ifdef DEBUG_ATOP
  printf("\nPER: L-J indexes:\n");
#endif
for(i=0; i<atop->natom; i++)
  {
    fscanf(ambtop, " %d", atop->iacper+i);
#ifdef DEBUG_ATOP
    printf("%6d:%10d\n", i, atop->iacper[i]);
#endif
  }


/* PERTURBED CHARGES */
#ifdef DEBUG_ATOP
  printf("\nPER: charges:\n");
#endif
atop->cgper = (double*) xmalloc((sizeof(double))*(atop->natom));
for(i=0; i < atop->natom; i++)
  {
    fscanf(ambtop, " %le", atop->cgper+i);
    atop->cgper[i] = atop->cgper[i]/sqrt(332.0000);  /* convert to electrons */
#ifdef DEBUG_ATOP
    printf("%6d:%20.8f\n", i, atop->cgper[i]); 
#endif
  }

  }
else
  {
#ifdef DEBUG_ATOP
  printf("\nno PERTURBATION info");
#endif
  }

/*  --------------------------- POLARIZATION ----------------------------- */
/* IPOL is not im AMBER topology documentation !!! */

atop->ipol = 0;

if (atop->ipol == 1)
  {
#ifdef DEBUG_ATOP
  printf("\nreading POLARIZATION info...\n");
#endif

/* PERTURBED POLARIZABILITIES */
#ifdef DEBUG_ATOP
  printf("\nPOL: atomic polarizabilities:\n");
#endif
atop->atpol = (double*) xmalloc((sizeof(double))*(atop->natom));
for(i=0; i < atop->natom; i++)
  {
    fscanf(ambtop, " %le", atop->atpol+i);
    atop->chrg[i] = atop->chrg[i]/sqrt(332.0000);  /* convert to electrons */
#ifdef DEBUG_ATOP
    printf("%6d:%20.8f\n", i, atop->atpol[i]); 
#endif
  }

  }
else
  {
#ifdef DEBUG_ATOP
  printf("\nno POLARIZATION info");
#endif
  }


/*  -------------------PERTURBATION + POLARIZATION --------------------------- */
if (atop->ipol == 1 && atop->ifpert == 1 )
  {
#ifdef DEBUG_ATOP
  printf("\nreading PER-POLARIZATION info...\n");
#endif

/* POLARIZABILITIES AT LAMBDA=1 */
#ifdef DEBUG_ATOP
  printf("\nPERT-POL: atomic polarizabilities:\n");
#endif
atop->atpol1 = (double*) xmalloc((sizeof(double))*(atop->natom));
for(i=0; i < atop->natom; i++)
  {
    fscanf(ambtop, " %le", atop->atpol1+i);
#ifdef DEBUG_ATOP
    printf("%6d:%20.8f\n", i, atop->atpol1[i]); 
#endif
  }

  }
else
  {
#ifdef DEBUG_ATOP
  printf("\nno PERT-POLARIZATION info");
#endif
  }


/* final message */
#ifdef DEBUG_ATOP
printf("\n\nReading of AMBER TOP finished\n\n");
#endif

/* ------------------------------------------------------------------ */
/* close file */
close_file(ambtop);

#ifdef DEBUG_ATOP
printf("\n\nFile AMBER TOP closed\n\n");
#endif

#ifdef DEBUG_MEM
  muntrace();
#endif

}

